#include <stdio.h>
#include <time.h>

typedef unsigned long long ull; 
typedef unsigned long  ul; 

main(int argc, char ** argv){

	/********** variables for storing user defined keys    ***********/

	ull K1, K2, K3, K4 ; 
	
	/*** Registers for 6 LFSRs, for each Ly and block D, n= yb +a***/
	ull 
		X1_bl0, X1_bl1, X1_bl2, X1_bl3, X1_bl4, D1,  
		X2_bl0, X2_bl1, X2_bl2, X2_bl3, X2_bl4, D2, 
		X3_bl0, X3_bl1, X3_bl2, X3_bl3, X3_bl4, D3, 
		X4_bl0, X4_bl1, X4_bl2, X4_bl3, X4_bl4, D4, 
		X5_bl0, X5_bl1, X5_bl2, X5_bl3, X5_bl4, D5, 
		X6_bl0, X6_bl1, X6_bl2, X6_bl3, X6_bl4, D6 ;    

	/***W_{k+y} corresponding to each LFSR ***********/ 

	ull  newbit1, newbit2, newbit3, newbit4, newbit5, newbit6; 	  
 
	/******* Output block W_k coreesponding to LFSR Ly *******/ 
 
	ull  X1, X2 , X3 , X4, X5, X6 ;  
	ull  temp1, temp2, temp3, temp4, temp5 , temp6 ;  

	ull  f,G,H,F, 				/* for Boolean function */	
	     ALLONE= 18446744073709551615; 	/* AllONE = 2^64-1 */

	ul   loop=0, 
	     blNo = 1024 * 1024 * 1024 ;       /* 2^30 */

	clock_t t1, t2 ;		       /* for time measurement */
	double tm, speed ; 

	FILE *fp;			      /* for file reading */   		
	
	/**************Key Scheduling Algorithm***********************/
	
	if ((fp = fopen (argv[1], "rb") )== NULL){
	
		printf ("Key file not found\n") ;
		exit () ;	
		
	}
	fscanf (fp, "%Lu%Lu", &K1, &K2);  /*reads the key */
	K3 = K1 ^ ALLONE ;
	K4 = K2 ^ ALLONE ;	
	
	X1_bl0 = K1 ; X1_bl1 = K2; X1_bl2 = K3 ; X1_bl3 = K4; 
	X1_bl4 = K2 ^ ALLONE  ; D1 = K3 >> 52 ;  
	
	X2_bl0 = K4 ^ ALLONE ; X2_bl1 = K1 ^ ALLONE; X2_bl2 = K3 ^ ALLONE ; 
	X2_bl3 = K4 ^ ALLONE; X2_bl4 = K1 ; D2 = (K2 ^ ALLONE ) >> 51 ;  
	
	X3_bl0 = K4 ^ ALLONE ; X3_bl1 = K1; X3_bl2 = K2 ; X3_bl3 = K3 ^ ALLONE; 
	X3_bl4 = K1 ^ ALLONE  ; D3 = (K4 ^ ALLONE) >> 47 ;  
	
	X4_bl0 = K3 ^ ALLONE ; X4_bl1 = K2; X4_bl2 = K4 ; X4_bl3 = K3 ^ ALLONE; 
	X4_bl4 = K2 ; D4 = (K1 ^ ALLONE) >> 41;  
	
	X5_bl0 = K3 ; X5_bl1 = K2; X5_bl2 = K1 ^ ALLONE ; X5_bl3 = K4 ^ ALLONE; 
	X5_bl4 = K2 ^ ALLONE  ; D5 = K1 >> 31;  
	
	X6_bl0 = K4 ; X6_bl1 = K3; X6_bl2 = K3 ; X6_bl3 = K2 ^ ALLONE; 
	X6_bl4 = K4 ^ ALLONE  ; D6 = K1 >> 7;  

	/*********************Key generation starts ********************/
	printf ("Generating %lu number of blocks, each of 64 bits \n", blNo);

	t1 = clock () ;
		
	while (loop < blNo ){			
	
	  /******x^332 + x^123 + x^0 (a= 12, y=5, r0= 0, r1= 59)********/ 
	  /** LFSR registers [X1_bl4, X1_bl3, X1_bl2, X1_bl1, X1_bl0] **/

		temp1 = X1_bl0 ^ (X1_bl1 >> 59) ^ (X1_bl2 << 5);  /* W'_k */

		newbit1 = (temp1 << 12) ^ D1 ;			  /* W_{k+d} */	

		D1 =  temp1 >> 52  ; 				  /* b-a = 52 */
				  
		X1 = X1_bl0 ;					/* output block */ 
 
		/**** shifting registers by one block ******/ 

		X1_bl0 = X1_bl1 ;
		X1_bl1 = X1_bl2 ;
		X1_bl2 = X1_bl3 ;
		X1_bl3 = X1_bl4 ;
		X1_bl4 = newbit1 ; 
  
	  /******x^333 + x^2 + x^0 (a= 13, y=5, r0= 0, r1= 2)********/ 
	  /* LFSR registers [X2_bl4, X2_bl3, X2_bl2, X2_bl1, X2_bl0] **/ 
 
		temp2 = X2_bl0 ^ (X2_bl0 >> 2) ^ (X2_bl1 << 62);   
 
		newbit2 = (temp2 << 13) ^ D2 ;  
 
		D2 =  temp2 >> 51  ; 				/* b-a = 51 */ 
		X2 = X2_bl0 ;					/* output block */ 
 
		/**** shifting registers by one block ******/ 
 
		X2_bl0 = X2_bl1 ; 
		X2_bl1 = X2_bl2 ; 
		X2_bl2 = X2_bl3 ; 
		X2_bl3 = X2_bl4 ; 
		X2_bl4 = newbit2 ;  
 
	  /******x^337 + x^135 + x^0 (a= 17, y=5, r0= 0, r1= 7 )******/ 
	  /* LFSR registers [X3_bl4, X3_bl3, X3_bl2, X3_bl1, X3_bl0] **/ 
 
		temp3 = X3_bl0 ^ (X3_bl2 >> 7 ) ^ (X3_bl3 << 57 );   
 
		newbit3 = (temp3 << 17 ) ^ D3 ;  
 
		D3 =  temp3 >> 47  ; 			/* b-a = 47 */ 
				
		X3 = X3_bl0 ;			       /* output block */ 
 
		/**** shifting registers by one block ******/ 
 
		X3_bl0 = X3_bl1 ; 
		X3_bl1 = X3_bl2 ; 
		X3_bl2 = X3_bl3 ; 
		X3_bl3 = X3_bl4 ; 
		X3_bl4 = newbit3 ;  
 
	  /******x^343 + x^205 + x^0 (a= 23, y=5, r0= 0, r1= 13 )******/ 
	  /* LFSR registers [X4_bl4, X4_bl3, X4_bl2, X4_bl1, X4_bl0] **/ 
 
		temp4 = X4_bl0 ^ (X4_bl3 >> 13 ) ^ (X4_bl4 << 51 );   
 
		newbit4 = (temp4 << 23) ^ D4 ;  
 
		D4 =  temp4 >> 41  ; 			/* b-a = 41 */ 
				   
		X4 = X4_bl0 ;				/* output block */ 
 
		/**** shifting registers by one block ******/ 
 
		X4_bl0 = X4_bl1  ; 
		X4_bl1 = X4_bl2  ; 
		X4_bl2 = X4_bl3  ; 
		X4_bl3 = X4_bl4  ; 
		X4_bl4 = newbit4 ;  
 
	  /******x^353 + x^69 + x^0 (a= 33, y=5, r0= 0, r1= 5 )******/ 
	  /* LFSR registers [D5, X5_bl4, X5_bl3, X5_bl2, X5_bl1, X5_bl0] **/ 
 
		temp5 = X5_bl0 ^ (X5_bl1 >> 5 ) ^ (X5_bl2 << 59 );   
 
		newbit5 = (temp5 << 33) ^ D5 ;  
 
		D5 =  temp5 >> 31  ; 			/* b-a = 31 */ 
				   
		X5 = X5_bl0 ;				/* output block */ 
 
		/**** shifting registers by one block ******/ 
 
		X5_bl0 = X5_bl1 ; 
		X5_bl1 = X5_bl2 ; 
		X5_bl2 = X5_bl3 ; 
		X5_bl3 = X5_bl4 ; 
		X5_bl4 = newbit5 ;  
	 
	  /******x^377 + x^75 + x^0 (a= 57, y=5, r0= 0, r1= 11 )******/ 
	  /* LFSR registers [D6, X6_bl4, X6_bl3, X6_bl2, X6_bl1, X6_bl0] **/ 
 
		temp6 = X6_bl0 ^ (X6_bl1 >> 11 ) ^ (X6_bl2 << 53 );   
 
		newbit6 = (temp6 << 57 ) ^ D6 ;  
 
		D6 =  temp6 >> 7  ; 		/* b-a = 7 */ 
				   
		X6 = X6_bl0 ;			/* output block */ 
 
		/**** shifting registers by one block ******/ 
 
		X6_bl0 = X6_bl1 ; 
		X6_bl1 = X6_bl2 ; 
		X6_bl2 = X6_bl3 ; 
		X6_bl3 = X6_bl4 ; 
		X6_bl4 = newbit6 ;  
		
	  /***********Boolean function***************/ 
	
		f = (ALLONE ^ X3 ) & X1 ;
		f ^=  (X3  & X2)  ;		/* (1 + X_n)f_1 + X_nf_2 */
	
		F = X5 ^ X4 ^ f ;
	
		G = (ALLONE ^ X5 ^ X4) & X1 ;
		G ^=  (X5 ^ X4 ) & X2 ; 
		G ^=  X5 ^ X3 ;

		H = (ALLONE ^ X6 ) & F  ^ (X6 & G) ; 

	/******************************************/
		
		loop ++ ;

	}
	
	t2 = clock() ;
	
	tm = ((double)(t2 - t1))/((double) CLOCKS_PER_SEC);
	printf ("Time taken = %10.3lf seconds\n", tm) ;
	speed = (((double)blNo) * 64.0)/ (tm * (1024.0* 1024.0 * 1024.0 )) ;
	printf ("Speed = %lf Giga bits per sec.\n", speed);
	
	
}

